/**
 * Package that contains experimental implementation of
 * "Binary-Encoded JSON-Like" data format handlers (parser,
 * generator, factory produce both, supporting constants).
 *<p>
 * See <a href="http://wiki.fasterxml.com/SmileFormat">Smile format specification</a> for more details.
 * 
 * @since 1.6
 */
package org.codehaus.jackson.smile;
