/**
 * Jackson-based JAX-RS provider that can automatically
 * serialize and deserialize resources for 
 * JSON content type (MediaType).
 *<p>
 * Also continues supporting functionality, such as
 * exception mappers that can simplify handling of
 * error conditions.
 */
package org.codehaus.jackson.jaxrs;
