/**
 * Package that contains concrete implementations of
 * {@link org.codehaus.jackson.type.JavaType}, as
 * well as the factory ({@link org.codehaus.jackson.map.type.TypeFactory}) for
 * constructing instances from various input data types
 * (like {@link java.lang.Class}, {@link java.lang.reflect.Type})
 * and programmatically (for structured types, arrays,
 * {@link java.util.List}s and {@link java.util.Map}s).
 */
package org.codehaus.jackson.map.type;
